/*****************************************************************************
ETHAN 1.0 

Automated detection of spherical particles from electron micrographs
Copyright (C) 2000 Teemu Kivioja (Teemu.Kivioja@cs.Helsinki.FI)

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.
*****************************************************************************/

#include "ethan.h"


/**************************************************************************/
/* find the maximum of a square                                           */
/**************************************************************************/
struct peak find_max(int ix, int iy, int lx, int ly, int **image)
{
    int i, j;
    struct peak max;    

    max.x = ix;
    max.y = iy;
    max.height = 0;
    for (j = iy; j < ly; j++) {
       for (i = ix; i < lx; i++) {
           if (image[j][i] > max.height) {
               max.x = i;
               max.y = j;
               max.height = image[j][i];
	   } 
       } 
    }

    return max;
}



/****************************************************************************/
/*  go through one square row and find the maximum values of the squares    */
/****************************************************************************/
void find_peaks(int square_width, int first_x, int first_y, 
                int dx, int dy, int **filtered_image, struct peak *peaks)
{
    int i, j, k = 0, last_x, last_y;
      
    for (j = 0; j * square_width < dy; j++) {
        if (j * square_width  + square_width >= dy)
            last_y = first_y + dy;           /* last square in y direction */
        else
            last_y = first_y + j * square_width + square_width;
        for (i = 0; i * square_width < dx; i++) {
            if (i * square_width  + square_width >= dx)
                last_x = first_x + dx;       /* last square in x direction */ 
            else
                last_x = first_x + i * square_width + square_width;
            peaks[k] = find_max(first_x + i * square_width, 
                                first_y + j * square_width, last_x, 
                                last_y, filtered_image); 
            k++;    
	}
    }
}



/***************************************************************************/
/* tell if two peaks are too close                                         */
/***************************************************************************/
int too_close(struct peak a, struct peak b, int limit)
{
    if (abs(a.x - b.x)*abs(a.x - b.x) +  abs(a.y - b.y)*abs(a.y - b.y)
        < limit*limit)
        return YES;
    else
        return NO;  
}



/****************************************************************************/
/* compare two peaks and if they are too close mark deleted the smaller one */
/****************************************************************************/
void edit_peak_pair(int k1, int i2, int j2, int n_squares_x, 
                    int n_squares_y, int limit, struct peak *peaks)
{
    int k2; 

    /* compare if second square exists and it's peak has not been deleted */
    if (i2 < n_squares_x  && j2 < n_squares_y) { 
        k2 = j2 * n_squares_x + i2;
        if (peaks[k2].x != DELETED) {               
             if (too_close(peaks[k1], peaks[k2], limit)) {
                if (peaks[k2].height > peaks[k1].height) {
	            peaks[k1].x = DELETED;
                } else {
                    peaks[k2].x = DELETED;
                }   
	     }
        }
    }
}           



/**************************************************************************/
/* compare two peaks and if they are too close mark both deleted          */
/**************************************************************************/
void delete_peak_pair(int k1, int i2, int j2, int n_squares_x, 
                      int n_squares_y, int limit, struct peak *peaks)
{
    int k2;
    
    /* compare if second square exists and it's peak has not been deleted */
    if (i2 < n_squares_x  && j2 < n_squares_y) { 
        k2 = j2 * n_squares_x + i2;
        if (peaks[k2].x != DELETED) {               
             if (too_close(peaks[k1], peaks[k2], limit)) {
                    peaks[k1].x = DELETED;
                    peaks[k2].x = DELETED;  
	     }
        }
    }
}           




/**************************************************************************/
/* mark deleted those peaks that are too close to other, higher,  peaks   */
/**************************************************************************/
void edit_peaks(int limit, int n_squares_x, int n_squares_y,
                  struct peak *peaks)
{
    int i, j, k, l; 
  
    for (j = 0; j < n_squares_y; j++)  
        for (i = 0; i < n_squares_x; i++) {
            k = j*n_squares_x + i;
            if (peaks[k].x == DELETED)    /* if the peak has already been  */  
                continue;                 /* deleted go to next square     */

            /* first row */
            for (l = i + 1; l <= i + 2; l++) {
                edit_peak_pair(k, l, j, n_squares_x, n_squares_y, 
                               limit, peaks);    
                if (peaks[k].x == DELETED)
                    continue;   
	    }
            if (peaks[k].x == DELETED)
                continue;   

            /* second row */
            for (l = i - 2; l <= i + 2; l++) {
                edit_peak_pair(k, l, j + 1, n_squares_x, n_squares_y, 
                               limit, peaks);    
                if (peaks[k].x == DELETED)
                    continue;   
	    }
            if (peaks[k].x == DELETED)
                continue;   

            /* third row */
            for (l = i - 1; l <= i + 1; l++) {
                edit_peak_pair(k, l, j + 2, n_squares_x, n_squares_y, 
                               limit, peaks);    
                if (peaks[k].x == DELETED)
                    continue;   
	    }
            if (peaks[k].x == DELETED)
                continue;   

	}
}


/**************************************************************************/
/* mark deleted those peak pairs which peaks are two close to each other  */
/**************************************************************************/
void delete_peaks(int limit, int n_squares_x, int n_squares_y,
                  struct peak *peaks)
{

    int i, j, k, l; 

    for (j = 0; j < n_squares_y; j++)  
        for (i = 0; i < n_squares_x; i++) {
            k = j*n_squares_x + i;
            if (peaks[k].x == DELETED)    /* if the peak has already been  */  
                continue;                 /* deleted go to next square     */

            /* first row */
            for (l = i + 1; l <= i + 3; l++) {
                delete_peak_pair(k, l, j, n_squares_x, n_squares_y, 
                               limit, peaks);    
                if (peaks[k].x == DELETED)
                    continue;   
	    }
            if (peaks[k].x == DELETED)
                continue;   

            /* second row */
            for (l = i - 3; l <= i + 3; l++) {
                delete_peak_pair(k, l, j + 1, n_squares_x, n_squares_y, 
                               limit, peaks);    
                if (peaks[k].x == DELETED)
                    continue;   
	    }
            if (peaks[k].x == DELETED)
                continue;   

            /* third row */
            for (l = i - 3; l <= i + 3; l++) {
                delete_peak_pair(k, l, j + 2, n_squares_x, n_squares_y, 
                               limit, peaks);    
                if (peaks[k].x == DELETED)
                    continue;   
	    }
            if (peaks[k].x == DELETED)
                continue;   

            /* fourth row */
            for (l = i - 2; l <= i + 2; l++) {
                delete_peak_pair(k, l, j + 3, n_squares_x, n_squares_y, 
                               limit, peaks);    
                if (peaks[k].x == DELETED)
                    continue;   
	    }
    }
}



/* If squares before k have  already been checked then peak in square k  */ 
/* can be too close to peaks in squares shown below                      */   
/*                                                                       */
/*                               |--|--|--|--|                           */
/*                               |k |  |  |  |                           */ 
/*                      |--|--|--|--|--|--|--|                           */
/*                      |  |  |  |  |  |  |  |                           */
/*                      |--|--|--|--|--|--|--|                           */
/*                      |  |  |  |  |  |  |  |                           */ 
/*                      |--|--|--|--|--|--|--|                           */
/*                         |  |  |  |  |  |                              */
/*                         |--|--|--|--|--|                              */




/***************************************************************************/
/* go through peak list and mark deleted those which are not high peaks    */
/***************************************************************************/
void check_peaks(struct peak *peaks, int k)
{
    int i, j = 0, avg = 0, height_limit;
    
    /* calculate the average height of the peak */
    for (i = 0; i < k; i++) {
        if ((peaks[i].x != DELETED)) {
            avg += peaks[i].height;
            j++;
        }    
    }
    if (j > 0) {
       avg = avg / j;
    }
    
  
    /* if the peak is too small compared to average height mark it deleted */
    height_limit = height_param * (avg - 12800) + 12800; 
    for (i = 0; i < k; i++) {
       if ((peaks[i].x != DELETED) && (peaks[i].height < height_limit) )
           peaks[i].x = DELETED; 
    }    
}



/****************************************************************************/
/* write particle area to a mrc file                                        */
/****************************************************************************/
void write_mrc_particle(char *filename, char *dir_name, FILE *fp, 
                        int header_size, int width, int height, struct peak p)
{
    int start_x, start_y;
    char name[500];
    unsigned char **piece, *rowp;
    char number_string[500];
    FILE *vfp;

    start_x = p.x - size / 2;
    start_y = p.y - size / 2;

    if (start_x >= 0 && start_x + size < width &&
        start_y >= 0 && start_y + size < height) {
  
            piece = (unsigned char **) allocate_2d(size, size, 1);
            if ((rowp = (unsigned char *) malloc(size)) == NULL) {
                fprintf(stderr, "Ethan: memory allocation failure, "); 
                fprintf(stderr, "write_mrc_particle\n");
                exit(ETHAN_MEMORY_ERROR);
            }    
            read_sampled_piece_c(filename, fp, header_size, width, height, 
                                 size, start_x, start_y, 
                                 size, size, 1, 1, piece, rowp);
            
            strcpy(name, dir_name);
            strcat(name, "/p_");
            sprintf(number_string, "%05d", file_number);
            strcat(name, number_string); 
            strcat(name, ".mrc");
            file_number++;

            vfp = mrc_write_open(name, size, size, 1); 
            fwrite(*piece, 1, size * size, vfp);
            if (ferror(vfp)) {
                fprintf(stderr, "Ethan: error writing file %s\n", 
                        name);
                exit(ETHAN_WRITE_ERROR);     
            }
            fclose(vfp);
            free_2d((void **) piece);
            free(rowp);

    } else {
        p.x = DELETED; 
    }  
}



/*************************************************************************/
/* refine the center of the particle with full resolution and write the  */
/* particle to a file if requested                                       */
/*************************************************************************/
void refine_and_cut(char *filename, char *dir_name, FILE *fp, int header_size, 
                    int width, int height, int sample_rate, int rb, int rs, 
                    struct peak *peaks, int k)
{
    unsigned char **piece, *rowp;
    int x, y, piece_width, piece_height, start_x, start_y, end_x, end_y;
    int i, best;
    int rbt[5], rst[5];
    int dist;
 
    dist = dist_param * rs;

    determine_radii(rb, rs, rbt, rst); 
    
    /* go through the peak list */
    for (i = 0; i < k; i++) {
        if (peaks[i].x != DELETED) {
	    
            x = sample_rate * peaks[i].x;
            y = sample_rate * peaks[i].y;
            best = peaks[i].best;
             

            start_x = (x < dist + rbt[best]) ?  0 : x - dist - rbt[best];
            start_y = (y < dist + rbt[best]) ?  0 : y - dist - rbt[best];

            end_x   = (x + dist + rbt[best] > width - 1) ? 
                      width - 1 : x + dist + rbt[best];
            end_y   = (y + dist + rbt[best] > height - 1) ? 
                      height - 1 : y + dist + rbt[best];

            piece_width  = end_x - start_x + 1;
            piece_height = end_y - start_y + 1;

            piece = (unsigned char **) 
                     allocate_2d(piece_height, piece_width, 1);
            if ((rowp = (unsigned char *) malloc(piece_width)) == NULL) {
                fprintf(stderr, "Ethan: memory allocation failure "); 
                fprintf(stderr, "last_check\n");
                exit(ETHAN_MEMORY_ERROR);
            }    
            read_sampled_piece_c(filename, fp, header_size, width, height, 
                                 piece_width, start_x, start_y, 
                                 piece_width, piece_height, 1, 1, piece, rowp);

            /* find the best center */
            peaks[i] = max_filter(rbt[best], rst[best], rbt[best], rbt[best], 
                                  piece_width - 2 * rbt[best], 
                                  piece_height - 2 * rbt[best], piece);

            peaks[i].x += start_x;
            peaks[i].y += start_y;      

            if (dir_name != NULL)              
                write_mrc_particle(filename, dir_name, fp, header_size, width, 
                                  height, peaks[i]);

            free_2d((void **) piece);
            free(rowp);                
        }
    }
}
























